/**
 *    This file is part of OtviAI.
 *
 *    OtviAI is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    OtviAI is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with OtviAI.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright 2008-2010 Otto Visser.
 * Suggestions/comments and bugs can go to: otviAI@otvi.nl
 */

/**
 * list of tested NEWGrf's:
 * - ECS
 * - HEQS (combination with ECS is borked)
 */

/**
 * TODO:
 - do the whole thing in a transaction to try if you can do it for the money you have now?
 - check roads to depots; always generates err_unknown: aka I'm ignoring those also when I shouldn't
 - search with increasing radius: subtract inner circle to prevent that to be checked 10 times
 - max route cost based on max height found on map
 - add service to order?
 - support replacing vehicles if autorenew isn't working
 - check: depot heigth == track height
 - improve train depot building
 - reuse of town train stations
 - seperate depot building from track building. Perhaps list of stations associated with depots again.
 - remove stations no longer in use
 - add boat support
 - BUG: building passenger train for oil well???
 - combination HEQS and ECS seems borked
 */

require("util.nut");
require("town.nut");

require("train.nut");
require("road.nut");

class OtviAI extends AIController {
	biggestTown = -1;	
	subsidyBuilt = false;	// TODO fix: value is incorrected in some cases
	
	mailCargoId = -1;

	MIN_DISTANCE = 50;	// ends at 40
	MAX_DISTANCE = 100;	// ends at 200

	Town = null;

	Train = null;
	Road = null;

	transports = null;

	uberstartTick = 0;

	function Save() {
		// TODO
		return {};
	}

	function Load(version, data) {
		Debug("TODO: implement load and save");
	}

	function constructor() {
		::INSTANCE <- this;
		local list = AICargoList();
		for (local i = list.Begin(); list.HasNext(); i = list.Next()) {
			if (AICargo.HasCargoClass(i, AICargo.CC_PASSENGERS) && AICargo.GetTownEffect(i) == AICargo.TE_PASSENGERS) {
				Debug("Found passenger cargo: " + AICargo.GetCargoLabel(i));
				::PASSENGER_ID <- i;
			}
		}
		local list = AICargoList();
		for (local i = list.Begin(); list.HasNext(); i = list.Next()) {
			if (AICargo.HasCargoClass(i, AICargo.CC_MAIL) && AICargo.GetTownEffect(i) == AICargo.TE_MAIL) {
				Debug("Found mail cargo: " + AICargo.GetCargoLabel(i));
				::MAIL_ID <- i;
			}
		}

		Town = OtviTown();
		transports  = [];

		if (AIGameSettings.IsDisabledVehicleType(AIVehicle.VT_RAIL)) {
			Warning("Trains are disabled!!!");
		} else {
			Train = OtviTrAIn();
			transports.append(Train);
		}

		if (AIGameSettings.IsDisabledVehicleType(AIVehicle.VT_ROAD)) {
			Warning("Road vehicles are disabled!!!");
		} else { 
			Road = OtviRoad();
			transports.append(Road);
		}

		if (AIGameSettings.IsDisabledVehicleType(AIVehicle.VT_WATER)) {
			Warning("Water transport is disabled!!!");
		} else { 
			Warning("Water transport is enabled, but not yet implemented.");
		}

		if (AIGameSettings.IsDisabledVehicleType(AIVehicle.VT_AIR)) {
			Warning("Air transport is disabled!!!");
		} else { 
			Warning("Air transport is enabled, but not yet implemented.");
		}

		mailCargoId = OtviUtil.getMailCargoID();

		local towns = AITownList();
		towns.Valuate(AITown.GetPopulation);
		towns.Sort(AIAbstractList.SORT_BY_VALUE, false);
		biggestTown = towns.Begin();
	}

	function Start() {
		if (!AICompany.SetName("OTVI")) {
			local i = 0;
			Town.setServed(biggestTown);	// prevent clusterfuck with other instance of OtviAI
			biggestTown = towns.Next();
			while (!AICompany.SetName("OTVI franchise #" + i)) {
				Town.setServed(biggestTown);	// prevent clusterfuck with other instance of OtviAI
				biggestTown = towns.Next();
				i++;
			}
		}

		if (transports.len() == 0) {
			Error("Can't build anything, going into a coma...");
			while(true) {
				manageLoan();
				Sleep(1024);
			}
		}
		
		local breakdownLevel = AIGameSettings.GetValue("difficulty.vehicle_breakdowns"); 
		if (AIGameSettings.IsValid("difficulty.vehicle_breakdowns") && breakdownLevel != 0) {
			Warning("Breakdowns are on; enabling replacement of vehicles...");
			if (!AICompany.SetAutoRenewStatus(true))
				Error("Couldn't enable auto renew; TODO: implement manual way");
			AICompany.SetAutoRenewMonths(breakdownLevel == 1 ? 12 : 6);
			AICompany.SetAutoRenewMoney(100000); //	TODO calculate!!!! this way it shouldn't interfere with normal building
		} else {
			AICompany.SetAutoRenewStatus(false);
		}

		local canOperate = false;
		foreach (transport in transports) {
			if (transport.isOperational())
				canOperate = true;
		}

		while(!canOperate) {
			foreach (transport in transports) {
				if (transport.isOperational())
					canOperate = true;
			}

			manageLoan();	// no loan (also no hq) for as long as possible
			checkEvents();	// wait for suitable equipment
			
			this.Sleep(1);
		}

		
		// the main control loop
		local curdate = AIDate.GetCurrentDate();
		local currentYear = AIDate.GetYear(curdate);
		local currentMonth = AIDate.GetMonth(curdate);
		local lastYear = currentYear;
		local endYear = currentYear + 10;	// last year to chance search radius for new town connections
		local lastExpansionMonth = currentMonth;	// TODO convert to daycount?
		local hqBuilt = false;
		
		while(true){
			if(currentYear != lastYear && currentYear < endYear) {
				lastYear = currentYear;
				MAX_DISTANCE += 10;
				MIN_DISTANCE -= 1;
			}
			
			local canEnhance = false;
			foreach (transport in transports) {
				if (transport.canEnhance())
					canEnhance = true;
			}

			transports.sort(OtviTransport.enhanceSort);
			if(AICompany.GetMaxLoanAmount() - AICompany.GetLoanAmount() + AICompany.GetBankBalance(AICompany.COMPANY_SELF) > transports[0].getEnhanceCost() && transports[0].canEnhance() && lastExpansionMonth != currentMonth) {
				if(lastMail1 != -1 && lastMail2 != -1) {
					// add buses and mail trucks where there is plenty cargo
					local lorry1 = AIStation.GetStationID(lastMail1);
					local lorry2 = AIStation.GetStationID(lastMail2);
					if(min(AIStation.GetCargoWaiting(lorry1, mailCargoId), AIStation.GetCargoWaiting(lorry2, mailCargoId)) > 3 * AIVehicle.GetCapacity(lastMailVehicle, mailCargoId)) {
						AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
						Debug("Adding mailtrucks at " + AIStation.GetName(lorry1) + " to " + AIStation.GetName(lorry2));
						if (connectMailRoadStations(lastMail1, lastMail2))
							lastExpansionMonth = currentMonth;
					}
				}
				if(lastTown1 != -1 && lastTown2 != -1) {
					// add buses and mail trucks where there is plenty cargo
					local pass1 = AIStation.GetStationID(lastTown1);
					local pass2 = AIStation.GetStationID(lastTown2);
					if(min(AIStation.GetCargoWaiting(pass1, passengerCargoId), AIStation.GetCargoWaiting(pass2, passengerCargoId)) > 4 * AIVehicle.GetCapacity(lastPassengerVehicle, passengerCargoId) && lastMaxBusses) {
						AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
						Debug("Adding busses at " + AIStation.GetName(pass1) + " to " + AIStation.GetName(pass2));
						if(connectPassengerRoadStations(lastTown1, lastTown2)) {
							lastMaxBusses--;
							lastExpansionMonth = currentMonth;
						}
					}
				}
			}
			
			transports.sort(OtviTransport.expansionSort);

			// TEMP
/*
			local towns = AITownList();
			towns.Valuate(AITown.GetPopulation);
			towns.Sort(AIAbstractList.SORT_BY_VALUE, false);
			biggestTown = towns.Begin();
			while (towns.HasNext()) {
				Train.BuildTrainStation(biggestTown);
				biggestTown = towns.Next();
			}*/
			// END TEMP

			if(AICompany.GetMaxLoanAmount() - AICompany.GetLoanAmount() + AICompany.GetBankBalance(AICompany.COMPANY_SELF) > transports[0].getExpansionCost()) {
				Debug("We've got enough money to connect stuffs!");
				if(AICompany.GetBankBalance(AICompany.COMPANY_SELF) < 125000)
					AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
				//Debug("sub count: " + AISubsidyList().Count() + "; building one: " + subsidyBuilt);
				if (!connect2Towns()) {
					manageLoan();
					Sleep(100);
				}
			}
						
			if (!hqBuilt) {
				buildHQ(biggestTown);
				hqBuilt = true;
			}

			manageLoan();
			
			Sleep(1);
			
			checkEvents();

			curdate = AIDate.GetCurrentDate();
			currentYear = AIDate.GetYear(curdate);
			currentMonth = AIDate.GetMonth(curdate);
		}
	}

	
	// 6.3 tiles per day for passengers? // TODO: check or calculate
	function incomeSort(targetTown, startTile, cargoID) {
		local distance = AIMap.DistanceManhattan(startTile, AITown.GetLocation(targetTown));
		local days = distance/7 + 1;	//underestimate
		//Debug(AITown.GetName(targetTown) + " at distance: " + distance + ", days: " + days + " income: " + AICargo.GetCargoIncome(cargoID, distance, days) + " maxProd: " + AITown.GetMaxProduction(targetTown, cargoID));
		return AICargo.GetCargoIncome(cargoID, distance, days) * (AITown.GetLastMonthProduction(targetTown, cargoID) - AITown.GetLastMonthTransported(targetTown, cargoID)) * (365/days);
	}
	
	function checkEvents() {
		while (AIEventController.IsEventWaiting()) {
			local e = AIEventController.GetNextEvent();
			//Debug("Got an eventA: " + e);
			switch (e.GetEventType()) {
				case AIEvent.AI_ET_VEHICLE_CRASHED:
					local ec = AIEventVehicleCrashed.Convert(e);
					local v  = ec.GetVehicleID();
					Debug("We have a crashed vehicle (" + v + ")");
					// TODO: Handle the crashed vehicle; aka: clone?
					break;
				case AIEvent.AI_ET_ENGINE_AVAILABLE:
					Debug("New engine available; rechecking current engines...");
					foreach (transport in transports)
						transport.updateEngines();
					break;
				case AIEvent.AI_ET_VEHICLE_LOST:
					Debug("A vehicle got lost!!!");
					// TODO: nuke vehicle??
					break;
				case AIEvent.AI_ET_VEHICLE_WAITING_IN_DEPOT:
					Debug("A vehicle arrived in the depot for selling");
					local ec = AIEventVehicleWaitingInDepot.Convert(e);
					local v  = ec.GetVehicleID();
					AIVehicle.SellVehicle(v);
					break;
				case AIEvent. AI_ET_VEHICLE_UNPROFITABLE:
					Debug("Unprofitable vehicle detected; sending to depot");
					// TODO: evaluate on one more year?
					local ec = AIEventVehicleUnprofitable.Convert(e);
					local v  = ec.GetVehicleID();
					AIVehicle.SendVehicleToDepot(v);
					break;
				case AIEvent.AI_ET_SUBSIDY_OFFER:
					local ec = AIEventSubsidyOffer.Convert(e);
					local s = ec.GetSubsidyID();
					Debug("New subsidy detected: " + s);
					if (AICompany.GetMaxLoanAmount() - AICompany.GetLoanAmount() + AICompany.GetBankBalance(AICompany.COMPANY_SELF) > 70000) {
						Debug("We've got enough money to get the subsidized route!");
						if (AICompany.GetBankBalance(AICompany.COMPANY_SELF) < 125000)
							AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
						connectSubsidy(s);
					} else {
						Warning("Not enough money to get this one now.");
					}
					break;
				case AIEvent.AI_ET_SUBSIDY_AWARDED:
					subsidyBuilt = false;
					local ec = AIEventSubsidyAwarded.Convert(e);
					local s = ec.GetSubsidyID();
					if (AICompany.IsMine(AISubsidy.GetAwardedTo(s)))
						Debug("We got the subsidy!!!!");
					else
						Warning("!!! We failed to win the subsidy!!!!");
					break;
				case AIEvent.AI_ET_SUBSIDY_OFFER_EXPIRED:
					subsidyBuilt = false;
					Debug("A subsidy expired");
					break;
				default:
					//Debug("Unhandled event:" + e);
			}
		}
	}
	
	function manageLoan() {
		//repay loan
		local balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
		local balanceMod = (balance / AICompany.GetLoanInterval()) * AICompany.GetLoanInterval();
		local loan = AICompany.GetLoanAmount();
		
		if (balance < 0) {
			// not good... we lost money?
			AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
			balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
			balanceMod = (balance / AICompany.GetLoanInterval()) * AICompany.GetLoanInterval();
			loan = AICompany.GetLoanAmount();
		}
		
		if (loan > 0 && balanceMod > 0) {
			if (balanceMod > 100000)
				balanceMod = 100000;
			AICompany.SetLoanAmount(loan - balanceMod);
		}
	}

	function buildHQ(town) {
		Debug("building HQ...");
		// Find empty 2x2 square as close to town centre as possible
		local spotFound = false;
		local curRange = 1;
		local maxRange = sqrt(AITown.GetPopulation(town)/100) + 5; //TODO check value correctness
		local area = AITileList();
		
		while (curRange < maxRange) {
			area.AddRectangle(AITown.GetLocation(town) - AIMap.GetTileIndex(curRange, curRange), AITown.GetLocation(town) + AIMap.GetTileIndex(curRange, curRange));
			area.Valuate(AITile.IsBuildableRectangle, 2, 2);
			area.KeepValue(1);
			if (area.Count()) {
				for (local t = area.Begin(); area.HasNext(); t = area.Next()) {
					local opening = OtviRoad.getRoadTile(t);
					if(opening) {
						if (AICompany.BuildCompanyHQ(t)) {
							// TODO different name if franchise
							AISign.BuildSign(t, ("HQ OTVI"));
							return t;
						}
					}
				}
				curRange++;
			} else {
				curRange++;
			}
		}
		Debug("No possible HQ location found");
		return null;
	}

	function connect2Towns() {
		// look for biggest towns
		local townlist = AITownList();
		townlist.Valuate(Town.isServed);
		townlist.KeepValue(0);
		townlist.Valuate(AITown.GetLastMonthProduction, mailCargoId);
		townlist.Sort(AIAbstractList.SORT_BY_VALUE, false);
		if (townlist.IsEmpty()) {
			Debug("No more towns left");
			// TODO: open up possibilities somehow?
			Town.cleanUp();
			manageLoan();
			Sleep(100);
			return;
		}
		biggestTown = townlist.Begin();
		
		Debug("biggest town: " + AITown.GetName(biggestTown));
		local bigLoc = AITown.GetLocation(biggestTown);
		// find connecting town between MIN and MAX manhattan distance
		townlist = AITownList();
		townlist.Valuate(incomeSort, bigLoc, mailCargoId);
		townlist.Sort(AIAbstractList.SORT_BY_VALUE, false);
		local connectTown = townlist.Begin();
		local distance = AITile.GetDistanceManhattanToTile(bigLoc, AITown.GetLocation(connectTown));
		while (distance > MAX_DISTANCE || distance < MIN_DISTANCE) {
			//Debug("skipping town: " + AITown.GetName(connectTown) + " at distance " + AITile.GetDistanceManhattanToTile(bigLoc, AITown.GetLocation(connectTown)) + " value: " + townlist.GetValue(connectTown));
			if (!townlist.HasNext()) {	// TODO: add check like: OR skipped 20 or value < maxvalue/4 or something
				Debug("Out of towns");
				return false; // too bad, out of towns
			}
			connectTown = townlist.Next();
			distance = AITile.GetDistanceManhattanToTile(bigLoc, AITown.GetLocation(connectTown));
		}
		
		Debug("Connecting town: " + AITown.GetName(connectTown) + " at a manhattan distance of " + AITile.GetDistanceManhattanToTile(bigLoc, AITown.GetLocation(connectTown)) + " value: " + townlist.GetValue(connectTown));

		::startLoc <- biggestTown;
		::endLoc <- connectTown;
		transports.sort(OtviTransport.townTransportSort);// : (biggestTown, connectTown);
		local connected = false;
		local i = 0;
		while (!connected && i < transports.len()) {
			connected = transports[i].connect2Towns(biggestTown, connectTown);
			i++;
		}
		if (connected) {
			Debug("Connected.");
			Town.setServed(biggestTown);
		} else {
			Warning("Could not connect town with any available transport means; marking town...");
			Town.setServed(-1);
		}

		/*if (AITown.GetPopulation(biggestTown) < 1000 || AITown.GetPopulation(connectTown) < 1000)
			return connect2TownsByRoad(biggestTown, connectTown);
		else {
			if(connect2TownsByRail(biggestTown, connectTown) == false) {
				Warning("Train failed for this route, trying busses instead...");
				return connect2TownsByRoad(biggestTown, connectTown);	// try road if rail fails
			}
		}*/
	}

	function connectSubsidy(subID) {		
		if (!AISubsidy.IsValidSubsidy(subID)) {
			Error("Invalid subsidy detected");
			return false;
		}
		if (AISubsidy.GetSourceType(subID) == AISubsidy.SPT_TOWN) {
			local town1 = AISubsidy.GetSourceIndex(subID);
			local town2 = AISubsidy.GetDestinationIndex(subID);
			Debug("Connecting: " + AITown.GetName(town1) + " to: " + AITown.GetName(town2));
			::startLoc <- town1;
			::endLoc <- town2;
			transports.sort(OtviTransport.townTransportSort, town1, town2);
			local connected = false;
			local i = 0;
			while (!connected && i < transports.len()) {
				connected = transports[i].connect2Towns(town1, town2);
				i++;
			}
			if (connected) {
				Debug("Connected.");
			} else {
				Warning("Could not connect town with any available transport means; marking town...");
			}
		} else if (AISubsidy.GetSourceType(subID) == AISubsidy.SPT_INDUSTRY) {
			local loc1 = AIIndustry.GetLocation(AISubsidy.GetSourceIndex(subID));
			local loc2;
			if (AISubsidy.GetDestinationType(subID) == AISubsidy.SPT_TOWN)
				loc2 = AITown.GetLocation(AISubsidy.GetDestinationIndex(subID));
			else
				loc2 = AIIndustry.GetLocation(AISubsidy.GetDestinationIndex(subID));
			local distance = AITile.GetDistanceManhattanToTile(loc1, loc2);
			Debug("Route between 2 industries is " + distance);
			local subCargo = AISubsidy.GetCargoType(subID);

			::startIndustry <- AISubsidy.GetSourceIndex(subID);
			::endLoc <- loc2;
			::cargo_ID <- subCargo;
			transports.sort(OtviTransport.industryTransportSort);
			local connected = false;
			local i = 0;
			while (!connected && i < transports.len()) {
				connected = transports[i].connectSubIndustries(subID);
				i++;
			}
			if (connected) {
				Debug("Connected.");
			} else {
				Warning("Could not connect subsidy");
			}
		} else {
			Error("Invalid subsidy detected: " + subID);
		}
	}
}

