/**
 *    This file is part of OtviAI.
 *
 *    OtviAI is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    OtviAI is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with OtviAI.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright 2008-2010 Otto Visser.
 * Suggestions/comments and bugs can go to: otviAI@otvi.nl
 */

require("util.nut");
require("RailFinder/RailFinder.nut");

class OtviTrAIn {
	trainEngine = -1;
	passengerCart = -1;
	mailCart = -1;

	railFinder = null;
	COST_RAIL = 500;
	stationRadius = -1;

	myRailStationTownList = null;
	myRailStationLengthList = null;
	myStationsTileList = null;

	current_rail_type = 0;

	function constructor() {
		Debug("Initializing train transportation...");
		railFinder = RailMaker();

		railFinder.cost.tile = 1;
		railFinder.cost.max_cost = 30000;	// TODO: still an estimate
		railFinder.cost.turn = COST_RAIL;		// The cost that is added to _cost_tile if the direction changes.
		railFinder.cost.diagonal_tile = 0;
		railFinder.cost.new_rail = COST_RAIL;		// was: 5
		railFinder.cost.road_tile = 1;
		railFinder.cost.slope =	COST_RAIL * 2;		// The extra cost if a rail tile is sloped.
		railFinder.cost.bridge_per_tile = COST_RAIL * 4;	// The cost per tile of a new bridge, this is added to _cost_tile.
		railFinder.cost.tunnel_per_tile = COST_RAIL * 4;	// The cost per tile of a new tunnel, this is added to _cost_tile.
		railFinder.cost.coast =	COST_RAIL;		// The extra cost for a coast tile.
		railFinder.cost.max_bridge_length = 10;	// The maximum length of a bridge that will be build.
		railFinder.cost.max_tunnel_length = 10;	// The maximum length of a tunnel that will be build.

		local railList = AIRailTypeList();
		Debug("Rail type count: " + railList.Count());
		current_rail_type = railList.Begin();	// TODO: detect change and perhaps convert rail
		AIRail.SetCurrentRailType(current_rail_type);
		trainEngine = getEngine(current_rail_type);
		passengerCart = getCart(::PASSENGER_ID, current_rail_type);
		mailCart = getCart(::MAIL_ID, current_rail_type);

		stationRadius = AIStation.GetCoverageRadius(AIStation.STATION_TRAIN);

		// init town list
		myRailStationTownList = AITownList();
		myRailStationLengthList = AITownList();
		myStationsTileList = AIList();
	}

	function isOperational() {
		if (trainEngine == -1)
			return false;
		return true;
	}

	function getExpansionCost(distance = 100) {
		return 80000;	// TODO
	}

	function canEnhance() {
		return false;
	}

	function getTownTransportCost(startLoc, endLoc) {
		if (passengerCart == -1) {
			Debug("train returns 255");
			return 255;
		}

		local ret = 1;
		if (AITown.GetPopulation(startLoc) < 1000 || AITown.GetPopulation(endLoc) < 1000)
			ret = 3;

		local dist = abs(sqrt(AIMap.DistanceSquare(AITown.GetLocation(startLoc), AITown.GetLocation(endLoc))));
		Debug("Train from " + AITown.GetName(startLoc) + " to " + AITown.GetName(endLoc) + " has a distance of " + dist);
		if (dist < 20 || dist > 250) {
			// too damn close for train or too damn far for current pathfinder
			Debug("train returns " + (ret * 8));
			return (ret * 8);
		}
		if (dist < 42 || dist > 150) {
			Debug("train returns " + (ret * 4));
			return (ret * 4);
		}
		Debug("train returns " + ret);
		return ret;
	}

	function getIndustryTransportCost(start, endLoc, cargoID) {
		if (getCart(cargoID, current_rail_type) == -1)
			return 255;

		// TODO:something with production levels
		local dist = AIMap.DistanceSquare(AIIndustry.GetLocation(start), endLoc);
		if (dist < 20 || dist > 250) {
			// too damn close for train or too damn far for current pathfinder
			return 9;
		}
		if (dist < 42 || dist > 150) {
			return 7;
		}
		else
			return 1;
	}

	function getEnhanceCost() {
		local expandMoneyMin = AIEngine.GetPrice(trainEngine);
		expandMoneyMin += 5 * AIEngine.GetPrice(passengerCart);
		expandMoneyMin += 2 * AIEngine.GetPrice(mailCart);
		return expandMoneyMin;
	}

	function getEngine(rail_type) {
		local trainEngine_list = AIEngineList(AIVehicle.VT_RAIL);
		trainEngine_list.Valuate(AIEngine.CanRunOnRail, rail_type);
		trainEngine_list.KeepValue(1);

		//trainEngine_list.Valuate(AIEngine.GetReliability);	// dependend on whether breakdowns are on!!
		trainEngine_list.Valuate(AIEngine.GetMaxSpeed);

		trainEngine_list.KeepTop(1);
		if (trainEngine_list.IsEmpty()) {
			return -1;
		}

		local trainEngine = trainEngine_list.Begin();
		Debug("Picked trainEngine: " + AIEngine.GetName(trainEngine));

		return trainEngine;
	}

	function getCart(cargoID, railType) {
		local cart_list = AIEngineList(AIVehicle.VT_RAIL);
		cart_list.Valuate(AIEngine.CanRefitCargo, cargoID);
		cart_list.KeepValue(1);

		cart_list.Valuate(AIEngine.IsWagon);
		cart_list.KeepValue(1);
		
		cart_list.Valuate(AIEngine.CanRunOnRail, railType);
		cart_list.KeepValue(1);

		cart_list.Valuate(AIEngine.GetCapacity)
		cart_list.KeepTop(1);
		
		if (cart_list.IsEmpty()) {
			Warning("No suitable cart for " + AICargo.GetCargoLabel(cargoID) + " available (yet?)!!");
			return -1;
		}

		local cart = cart_list.Begin();
		Debug("Picked cart for cargo " + AICargo.GetCargoLabel(cargoID) + ": " + AIEngine.GetName(cart));

		return cart;
	}

	function updateEngines() {
		local railList = AIRailTypeList();
		local new_type = railList.Begin();	// TODO
		if (new_type != current_rail_type) {
			Warning("New rail type!");
			/*for (local i = 0; i < 10; i++) {
			// TODO start index dependant on freeform edges!!!
			if (AIRail.ConvertRailType(AIMap.GetTileIndex(1, 1), AIMap.GetTileIndex(AIMap.GetMapSizeX() - 2, AIMap.GetMapSizeY() - 2 ), new_type)) {
			Debug("Rail conversion success");
			} else {
			Error("At least partial failure converting rail!");
			}
			::INSTANCE.Sleep(25);
			}*/
			current_rail_type = new_type;
		}
		AIRail.SetCurrentRailType(current_rail_type);
		trainEngine = getEngine(current_rail_type);
		passengerCart = getCart(::PASSENGER_ID, current_rail_type);
		mailCart = getCart(::MAIL_ID, current_rail_type);
	}

	function stationLocationValuator(tile, length, width) { // TODO check in 4 directions!!!
		local landTiles = AITileList();
		if (AITile.IsBuildableRectangle(tile, length + 1, width)) {
			landTiles.AddRectangle(tile, tile + AIMap.GetTileIndex(length - 1, width - 1));
			//landTiles.Valuate(AITile.GetSlope);
			//landTiles.KeepValue(AITile.SLOPE_FLAT);
			if (landTiles.Count() == length * width) {
				return AIRail.RAILTRACK_NE_SW;
			}
			// Warning("location only has limited flat tiles in NE_SW dir: " + landTiles.Count());
		}
		if (AITile.IsBuildableRectangle(tile, width, length + 1)) {
			landTiles.AddRectangle(tile, tile + AIMap.GetTileIndex(width - 1, length- 1));
			//landTiles.Valuate(AITile.GetSlope);
			//landTiles.KeepValue(AITile.SLOPE_FLAT);
			if (landTiles.Count() == length * width) {
				return AIRail.RAILTRACK_NW_SE;
			}
			// Warning("location only has limited flat tiles in NW_SE dir: " + landTiles.Count());
		}
		return 0;
	}
	
	function connectSubIndustries(subID) {
		// TODO: length station = f(distance, cartSize, production);
		local length1 = 4;
		local length2 = 4;
		local station2 = null;
		local build2 = false;
		local targetTown;
		local station1 = buildStation(AIIndustry.GetLocation(AISubsidy.GetSourceIndex(subID)), length1);
		if (!station1)
			return false; // bugger

		if (AISubsidy.GetDestinationType(subID) == AISubsidy.SPT_TOWN) {
			targetTown = AISubsidy.GetDestinationIndex(subID);
			station2 = getTrainStation(targetTown);
			if (!station2) {
				length2 = abs(sqrt(AITown.GetPopulation(targetTown)/100));
				station2 = BuildTrainStation(targetTown);
				if (!station2) {
					AITile.DemolishTile(station1);
					return false;
				}
				build2 = true;
			}
		} else {
			station2 = buildStation(AIIndustry.GetLocation(AISubsidy.GetDestinationIndex(subID)), length2);

			if (!station2)
				return false; // bugger
		}

		if (!connectStations(station1, length1, station2, length2)) {
			// check if route complete; build other orientation of depot
			Error("Incomplete route or unable to find suitable depot location");
			AITile.DemolishTile(station1);
			if (build2)
				AITile.DemolishTile(station2);
			return false;
		} else {
			if (build2)
				addTrainStation(targetTown, station2, length2);

			return connectIndustryStations(station1, station2, AISubsidy.GetCargoType(subID), length1);
		}
	}

	function connectSubTowns(subID) {
		biggestTown = AISubsidy.GetSourceIndex(subID);
		Debug("First subsidy town: " + AITown.GetName(biggestTown));
		local connectTown = AISubsidy.GetDestinationIndex(subID);
		Debug("Connecting subsidy town: " + AITown.GetName(connectTown) + " at a manhattan distance of " + AITile.GetDistanceManhattanToTile(AITown.GetLocation(biggestTown), AITown.GetLocation(connectTown)));

		return connect2Towns(connectTown, biggestTown);
	}

	/**
	 * 2 stations build, make rail
	 */
	function connect(starts, goals, ignored = [], skipDepot = false) {
		railFinder.cost.max_cost = 2 * COST_RAIL * AIMap.DistanceManhattan(starts[0][0], goals[0][0]);
		railFinder.InitializePath(starts, goals, ignored);
		local path = false;
		local i = 0;
		local err;
		// TODO: base on mountains?
		local iterations = 300 * AIMap.DistanceManhattan(starts[0][0], goals[0][0]); // if we pass this; something is wrong
		//local iterations = -1;
		local startTick = GetTick();
		// TODO check whether higher values helps finding faster
		Debug("pathfinding for max " + railFinder.cost.max_cost + " cost and iterations: " + iterations);
		path = railFinder.FindPath(iterations);
		Debug("Total ticks passed: " + (GetTick() - startTick));
		
		if (path == false) {
			Warning("!! No path found, too little iterations...");
			railFinder.reset();
			return -1;
		}
		if (path == null) {
			Warning("!! No path found, too expensive...");
			railFinder.reset();
			return -1;
		}

		Debug("path found for cost: " + path.GetCost() + ", building");

		local prev = null;
		local prevprev = null;
		local prev3 = null;
		local depotBuilt = false;
		local built = 0;
		local depotLoc = null;
		local depotOpening;
		local railCount = 0;

		while (path != null) {
			if (prevprev != null) {
				if (AIMap.DistanceManhattan(prev, path.GetTile()) > 1) {
					if (AITunnel.GetOtherTunnelEnd(prev) == path.GetTile()) {
						AITunnel.BuildTunnel(AIVehicle.VT_RAIL, prev);
					} else {
						local bridge_list = AIBridgeList_Length(AIMap.DistanceManhattan(path.GetTile(), prev) + 1);
						bridge_list.Valuate(AIBridge.GetMaxSpeed);
						bridge_list.Sort(AIAbstractList.SORT_BY_VALUE, false);
						AIBridge.BuildBridge(AIVehicle.VT_RAIL, bridge_list.Begin(), prev, path.GetTile());
					}
					prevprev = prev;
					prev = path.GetTile();
					path = path.GetParent();
				} else {
					if (!AIRail.BuildRail(prevprev, prev, path.GetTile())) {
						switch(AIError.GetLastError()) {
							case AIError.ERR_VEHICLE_IN_THE_WAY:
								Warning("Building rail failed temporary: vehicle in the way...");
								// try again till the stupid vehicle moved?
								local tries = 100;
								while(!AIRail.BuildRail(prevprev, prev, path.GetTile()) && tries--) {
									::INSTANCE.Sleep(10);
								}
								break;
							case AIRoad.ERR_ROAD_WORKS_IN_PROGRESS:
								Warning("Building rail failed: road works in progress");
								// good, someone else is building this, skip tile
								break;
							case AIError.ERR_ALREADY_BUILT:
								//Debug("Building rail failed: already built");
								// even better; someone else already built this; silent ignore
								if (AIRail.ConvertRailType(prevprev, prev, current_rail_type))
									Warning("Converted some rail to new type");
								break;
							case AIError.ERR_LAND_SLOPED_WRONG:
								Warning("Building rail failed: wrong land slope");
								// not good; redo from start or terraform?
								return connect(starts, goals, ignores);
							case AIError.ERR_AREA_NOT_CLEAR:
								Warning("Building rail failed: not clear, demolishing tile...");
								AISign.BuildSign(path.GetTile(), ("I'm on your land, demolishing your stuffs"));
								// demolish and retry?
								if (AITile.DemolishTile(path.GetTile())) {
									AIRail.BuildRail(prevprev, prev, path.GetTile());
								} else {
									return connect(starts, goals, ignored);
								}
								break;
							case AIRail.ERR_CROSSING_ON_ONEWAY_ROAD:
								Warning("Can't build crossing on one way road");
								return connect(starts, goals, ignores);
							case AIError.ERR_NOT_ENOUGH_CASH:
								Warning("Panic: Out of cash!!");
								return false;
							case AIError.ERR_OWNED_BY_ANOTHER_COMPANY:
								Warning("Owned by another company; retry...");
								return connect(starts, goals, ignored);
							default:
								Warning("Unknown error during road building " + AIError.GetLastErrorString());
								break;	// no clue what this is; silent ignore for now
						}
					}
				}

				// TODO FIXME
				/*if (railCount++ % 8 == 0) {
					AIRail.BuildSignal(prevprev, prev, AIRail.SIGNALTYPE_PBS_ONEWAY);
				}*/
			}
			if (!skipDepot && !depotBuilt && built++ > 3 && (prev3 != null) && (prevprev != null)) {
				if ((AIRail.GetRailTracks(prevprev) & AIRail.RAILTRACK_NE_SW) && (AIRail.GetRailTracks(prev) & AIRail.RAILTRACK_NE_SW) && (AIRail.GetRailTracks(prev3) & AIRail.RAILTRACK_NE_SW) ) {
					depotLoc = prevprev + AIMap.GetTileIndex(0,-1);
					depotOpening = prevprev;
					depotBuilt = AIRail.BuildRailDepot(depotLoc, prevprev);
					if (depotBuilt) {
						if (!(AIRail.BuildRailTrack(prevprev, AIRail.RAILTRACK_NW_NE) && AIRail.BuildRailTrack(prevprev, AIRail.RAILTRACK_NW_SW))) {
							depotBuilt = false;
							AITile.DemolishTile(depotLoc);
						}
					}
				}
				if ((AIRail.GetRailTracks(prevprev) & AIRail.RAILTRACK_NW_SE) && (AIRail.GetRailTracks(prev) & AIRail.RAILTRACK_NW_SE) && (AIRail.GetRailTracks(prev3) & AIRail.RAILTRACK_NW_SE)) {
					depotLoc = prevprev + AIMap.GetTileIndex(-1, 0);
					depotOpening = prevprev;
					depotBuilt = AIRail.BuildRailDepot(depotLoc, prevprev);
					if (depotBuilt) {
						if (!(AIRail.BuildRailTrack(prevprev, AIRail.RAILTRACK_NW_NE) && AIRail.BuildRailTrack(prevprev, AIRail.RAILTRACK_NE_SE))) {
							depotBuilt = false;
							AITile.DemolishTile(depotLoc);
						}
					}
				}
			}
			prev3 = prevprev;
			prevprev = prev;
			prev = path.GetTile();
			path = path.GetParent();
		}

		if (depotBuilt) {
			Debug("path done, depot built");
			connect([[depotOpening, depotLoc]], starts, [], true);
			connect([[depotOpening, depotLoc]], goals, [], true);
			return depotLoc;
		} 
		if (!skipDepot && !depotBuilt){
			Warning("Could not place depot, but path done");
			return -1;
		}

		return 0;
	}

	function connectIndustryStations(station1, station2, cargo, cartCount) {
		local depot = myStationsTileList.GetValue(station1);
		local train1 = AIVehicle.BuildVehicle(depot, trainEngine);
		if (!AIVehicle.IsValidVehicle(train1))
			return false;

		local cart;
		local cartType = getCart(cargo, current_rail_type);
		while (cartCount--) {
			cart = AIVehicle.BuildVehicle(depot, cartType);
			if (!AIVehicle.IsValidVehicle(cart)) {
				Error(AIError.GetLastErrorString());
			} else {
				AIVehicle.MoveWagon(cart, 0, train1, 0);
			}
		}

		Debug("Built train: " + train1);
		AIOrder.AppendOrder(train1, station1, AIOrder.AIOF_FULL_LOAD_ANY | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIOrder.AppendOrder(train1, station2, AIOrder.AIOF_NON_STOP_INTERMEDIATE | AIOrder.AIOF_SERVICE_IF_NEEDED);
		AIVehicle.StartStopVehicle(train1);

		return true;
	}

	function connectTownStations(station1, station2, cartCount) {
		if (cartCount > 12) {
			Warning("TODO: deploy more then 1 train");
			cartCount = 12;
		}

		local depot = myStationsTileList.GetValue(station1);
		local train1 = AIVehicle.BuildVehicle(depot, trainEngine);
		if (!AIVehicle.IsValidVehicle(train1)) {
			Error("Error building train engine");
			return false;
		}
		/*if(!AIVehicle.RefitVehicle(train1, ::PASSENGER_ID)) {
			Debug("Error refitting train: " + AIError.GetLastErrorString());
			return false;
		}*/
		local cart;
		local mailCarts = cartCount/4;
		if (cartCount + mailCarts > 12) {
			cartCount = 12 - mailCarts;
		}
		while (mailCarts--) {
			cart = AIVehicle.BuildVehicle(depot, mailCart);
			if (!AIVehicle.IsValidVehicle(cart)) {
				Error(AIError.GetLastErrorString());
			} else {
				AIVehicle.MoveWagon(cart, 0, train1, 0);
			}
		}
		while (cartCount--) {
			cart = AIVehicle.BuildVehicle(depot, passengerCart);
			if (!AIVehicle.IsValidVehicle(cart)) {
				Error(AIError.GetLastErrorString());
			} else {
				AIVehicle.MoveWagon(cart, 0, train1, 0);
			}
		}

		Debug("Built train: " + train1);
		//lastPassengerVehicle = train1;
		AIOrder.AppendOrder(train1, station1, AIOrder.AIOF_FULL_LOAD_ANY | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIOrder.AppendOrder(train1, station2, AIOrder.AIOF_FULL_LOAD_ANY | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIOrder.AppendOrder(train1, station2, AIOrder.AIOF_GOTO_NEAREST_DEPOT | AIOrder.AIOF_SERVICE_IF_NEEDED);
		AIVehicle.StartStopVehicle(train1);

		/* local train2 = AIVehicle.BuildVehicle(myStationsTileList.GetValue(station2), trainEngine);
		if (!AIVehicle.IsValidVehicle(train2))
			return false; 	// TODO change to return 1 for 1 success
		if(!AIVehicle.RefitVehicle(train2, ::PASSENGER_ID)) {
			Debug("Error refitting train: " + AIError.GetLastErrorString());
			return false; // TODO
		}
		Debug("Built train: " + train2);
		AIOrder.AppendOrder(train2, station1, AIOrder.AIOF_FULL_LOAD | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIOrder.AppendOrder(train2, station2, AIOrder.AIOF_FULL_LOAD | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIVehicle.StartStopVehicle(train2); */

		return true;	// TODO change to 2 
	}
	
	function buildStation(target, length) {
		// TODO: centre isn't always centre; for industry it's the top 
		Debug("Building station...");
		// Find empty block of squares as close to target as possible
		local spotFound = false;
		local curRange = 1;
		local maxRange = 8 + length; //TODO check value correctness (coverage + size of something?)
		if (length > 7)
			length = 7;
		local width = (length == 7 ? 2 : 1);

		// you need changing tracks for multiple platforms
		local totalLength = length + width;
		local area = AITileList();
		Debug("building station (lxb) " + length + " x " + width);

		while (curRange <= maxRange) {
			//Debug("looking with range " + curRange);
			area.AddRectangle(target + AIMap.GetTileIndex(curRange, curRange), target + AIMap.GetTileIndex(-curRange, -curRange));
			area.Valuate(stationLocationValuator, totalLength, width);
			area.RemoveValue(0);
			if (area.Count()) {
				Debug("Found " + area.Count() + " possible options");
				for (local t = area.Begin(); area.HasNext(); t = area.Next()) {
					if (area.GetValue(t) == AIRail.RAILTRACK_NE_SW) {
						if (!AITile.LevelTiles(t, t + AIMap.GetTileIndex(length + width, width)))
							Error(AIError.GetLastErrorString());
					} else if (area.GetValue(t) == AIRail.RAILTRACK_NW_SE) {
						if (!AITile.LevelTiles(t, t + AIMap.GetTileIndex(width, length + width)))
							Error(AIError.GetLastErrorString());
					}
					if (AIRail.BuildRailStation(t, area.GetValue(t), width, length, AIStation.STATION_JOIN_ADJACENT)) {
						// TODO: support for wider stations and having exits on the other side
						if (area.GetValue(t) == AIRail.RAILTRACK_NE_SW) {
							local loc = t + AIMap.GetTileIndex(totalLength - 1, 0);
							AIRail.BuildRailTrack(loc, AIRail.RAILTRACK_NE_SW);
							if (width == 1) {
								AIRail.BuildSignal(loc, loc + AIMap.GetTileIndex(-1, 0), AIRail.SIGNALTYPE_PBS);
							} else if (width == 2) {
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length, 0), AIRail.RAILTRACK_NE_SW);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length, 1), AIRail.RAILTRACK_NW_SW);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length, 0), AIRail.RAILTRACK_SW_SE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length, 0), AIRail.RAILTRACK_NE_SE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length, 1), AIRail.RAILTRACK_NE_SW);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length, 1), AIRail.RAILTRACK_NW_NE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length + 1, 1), AIRail.RAILTRACK_NE_SW);
								AIRail.BuildSignal(loc, loc + AIMap.GetTileIndex(-1, 0), AIRail.SIGNALTYPE_PBS_ONEWAY);
							} else {
								Error("Stations wider then 2 not yet supported");
							}								
						} else if (area.GetValue(t) == AIRail.RAILTRACK_NW_SE) {
							local loc = t + AIMap.GetTileIndex(0, totalLength - 1);
							AIRail.BuildRailTrack(loc, AIRail.RAILTRACK_NW_SE);
							if (width == 1) {
								AIRail.BuildSignal(loc, loc + AIMap.GetTileIndex(0, -1), AIRail.SIGNALTYPE_PBS);
							} else if (width == 2) {
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(0, length), AIRail.RAILTRACK_NW_SE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(1, length), AIRail.RAILTRACK_NW_SE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(0, length), AIRail.RAILTRACK_NW_SW);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(0, length), AIRail.RAILTRACK_SW_SE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(1, length), AIRail.RAILTRACK_NE_SE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(1, length), AIRail.RAILTRACK_NW_NE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(1, length + 1), AIRail.RAILTRACK_NW_SE);
								AIRail.BuildSignal(loc + AIMap.GetTileIndex(1, 0), loc + AIMap.GetTileIndex(1, -1), AIRail.SIGNALTYPE_PBS_ONEWAY);
							} else {
								Error("Stations wider then 2 not yet supported");
							}								
						}
						return t;
					}
					// TODO: if err == area not clear: demolish?
					//Debug("Error building passenger station " + AIError.GetLastErrorString());
					//AISign.BuildSign(t, AIError.GetLastErrorString());
				}
				curRange++;
			} else {
				curRange++;
			}
		}
		Debug("No possible station location found");
		return null;
	}
	
	function connect2Towns(biggestTown, connectTown) {
		local build1 = false;
		local build2 = false;

		local length1 = abs(sqrt(AITown.GetPopulation(biggestTown)/100));
		local station1 = getTrainStation(biggestTown);
		if (station1 != 0) {
			if (getStationLength(biggestTown) < length1) {
				Debug("Building a bigger and better station in " + AITown.GetName(biggestTown));
				station1 = BuildTrainStation(biggestTown);
				if (!station1) {
					Warning("!!! Building bigger station failed in: " + AITown.GetName(biggestTown));
					Debug("Using old existing station instead...");
					station1 = getTrainStation(biggestTown);
				} else {
					build1 = true;
				}
			}
		} else {
			station1 = BuildTrainStation(biggestTown);
			if (!station1) {
				Warning("!!! Building station failed in: " + AITown.GetName(biggestTown));
				return false;
			}
			build1 = true;
		}
		
		local length2 = abs(sqrt(AITown.GetPopulation(connectTown)/100));
		local station2 = getTrainStation(connectTown);
		if (station2 != 0) {                                                                                                             
			if (getStationLength(connectTown) < length2) {
				Debug("Building a bigger and better station in " + AITown.GetName(connectTown));
				station2 = BuildTrainStation(connectTown);
				if (!station2) {
					Warning("!!! Building bigger station failed in: " + AITown.GetName(connectTown));
					Debug("Using old existing station instead...");
					station2 = getTrainStation(connectTown);
				} else {
					build2 = true;
				}
			}
		} else {
			station2 = BuildTrainStation(connectTown);
			build2 = true;
		}

		if (!station2) {
			Warning("!!! Building station failed in: " + AITown.GetName(connectTown));
			Warning("Removing other station...");
			// TODO: hand to town and remove signal and rail as well
			if (build1) {
				AITile.DemolishTile(station1);
			}
			return false;
		}

		if (connectStations(station1, length1, station2, length2)) {
			//we picked BIG towns, lets try to build more carts in 1 go
			local pop1 = AITown.GetPopulation(biggestTown);
			local pop2 = AITown.GetPopulation(connectTown);
			// TODO: see whether this value max sense:
			local divider = AIEngine.GetCapacity(passengerCart) * 10;
			local maxCarts = (pop1 < pop2 ? pop1 : pop2) / divider;	// built a cart for every hundred passengers in the smallest town
			if (maxCarts < 1)
				maxCarts = 1;
			local maxLength = length1 > length2 ? (length1 * 2) - 2 : (length2 * 2) - 2;
			if (maxLength > 12)
				maxLength = 12;
			Debug("Want to deploy " + maxCarts + " carts for this route (pop)");
			if (maxCarts > maxLength) {
				Warning("Restricting train length from " + maxCarts + " to " + maxLength);
				maxCarts  = maxLength;
			}
			connectTownStations(station2, station1, maxCarts);
			if (build1)
				addTrainStation(biggestTown, station1, length1);
			if (build2)
				addTrainStation(connectTown, station2, length2);
			return true;
		} else {
			Warning("Failure... cleaning up....");
			// remove stuff we have build
			if (build1) {
				Warning("Cleaning station 1");
				local trains = AIVehicleList_Station(AIStation.GetStationID(station1));
				if (trains.IsEmpty())
					AITile.DemolishTile(station1);
			}
			if (build2) {
				Warning("Cleaning station 2");
				local trains = AIVehicleList_Station(AIStation.GetStationID(station2));
				if (trains.IsEmpty())
					AITile.DemolishTile(station2);
			}
			return false;
		}
	}

	// TODO support for bigger then 7x2
	function connectStations(station1, length1, station2, length2) {
		local extend1 = (length1 >= 7 ? 1 : 0);
		local extend2 = (length2 >= 7 ? 1 : 0);
		if (length1 > 7)
			length1 = 7;
		if (length2 > 7)
			length2 = 7;
		local openingStart;
		local openingGoal;
		local openingStart2;
		local openingGoal2;

		Debug("Route between 2 stations is " + AITile.GetDistanceManhattanToTile(station1, station2));

		if (AIRail.GetRailStationDirection(station1) == AIRail.RAILTRACK_NE_SW) {
			openingStart = station1 + AIMap.GetTileIndex(length1 + extend1 + 1, 0); // TODO FIXME
			openingStart2 = station1 + AIMap.GetTileIndex(length1 + extend1, 0); // TODO FIXME
		} else {
			openingStart = station1 + AIMap.GetTileIndex(0, length1 + extend1 + 1); // TODO FIXME
			openingStart2 = station1 + AIMap.GetTileIndex(0, length1 + extend1); // TODO FIXME
		}

		if (AIRail.GetRailStationDirection(station2) == AIRail.RAILTRACK_NE_SW) {
			openingGoal = station2 + AIMap.GetTileIndex(length2 + extend2 + 1, 0); // TODO FIXME
			openingGoal2 = station2 + AIMap.GetTileIndex(length2 + extend2, 0); // TODO FIXME generalyse
		} else {
			openingGoal = station2 + AIMap.GetTileIndex(0, length2 + extend2 + 1); // TODO FIXME
			openingGoal2 = station2 + AIMap.GetTileIndex(0, length2 + extend2); // TODO FIXME generalyse
		}
		/*AISign.BuildSign(openingStart, "F");
		AISign.BuildSign(openingGoal, "T");
		AISign.BuildSign(openingStart2, "F");
		AISign.BuildSign(openingGoal2, "T");*/
		local starts = [[openingStart, openingStart2]];
		local goals = [[openingGoal, openingGoal2]];
		local stations = [station1, station2];
		local skipDepot = false;
		if (AIRail.IsRailDepotTile(myStationsTileList.GetValue(station2))) {
			Debug("Reusing existing depot");
			skipDepot = true;
		}

		local depot = connect(starts, goals, stations, skipDepot);
		//if(!AIRail.IsRailDepotTile(depot)) {
			//depot = connect(starts, goals, stations);	// Try once more if it failed
		//}


		if (skipDepot) {
			if (depot == -1)
				return false;
			else
				depot = myStationsTileList.GetValue(station2);
		}

		if(!AIRail.IsRailDepotTile(depot)) {
			// check if route complete; build other orientation of depot
			Error("Incomplete route or unable to find suitable depot location");
			return false;
		} else {
			if (extend1 == 1) {
				if (AIRail.GetRailStationDirection(station1) == AIRail.RAILTRACK_NE_SW) {
					openingStart = station1 + AIMap.GetTileIndex(length1 + extend1 + 1, 1); // TODO FIXME
					openingStart2 = station1 + AIMap.GetTileIndex(length1 + extend1, 1); // TODO FIXME
				} else {
					openingStart = station1 + AIMap.GetTileIndex(1, length1 + extend1 + 1); // TODO FIXME
					openingStart2 = station1 + AIMap.GetTileIndex(1, length1 + extend1); // TODO FIXME
				}
				starts = [[openingStart, openingStart2]];
				// connect(starts, goals, stations, true);
			}
			if (extend2 == 1) {
				if (AIRail.GetRailStationDirection(station2) == AIRail.RAILTRACK_NE_SW) {
					openingGoal = station2 + AIMap.GetTileIndex(length2 + extend2 + 1, 1); // TODO FIXME
					openingGoal2 = station2 + AIMap.GetTileIndex(length2 + extend2, 1); // TODO FIXME generalyse
				} else {
					openingGoal = station2 + AIMap.GetTileIndex(1, length2 + extend2 + 1); // TODO FIXME
					openingGoal2 = station2 + AIMap.GetTileIndex(1, length2 + extend2); // TODO FIXME generalyse
				}
				goals = [[openingGoal, openingGoal2]];
				// connect(starts, goals, stations, true);
			}
			
			myStationsTileList.AddItem(station2, depot);

			// reverse track:

			if (AIRail.IsRailDepotTile(myStationsTileList.GetValue(station1)))
				skipDepot = true;
			else 
				skipDepot = false;
			depot = connect(goals, starts, stations, skipDepot);
			if(!skipDepot && AIRail.IsRailDepotTile(depot))
				myStationsTileList.AddItem(station1, depot);

			return true;
		}
	}

	function getTrainStation(town) {
		return myRailStationTownList.GetValue(town);
	}

	function getStationLength(town) {
		if (getTrainStation(town))
			return myRailStationLengthList.GetValue(town);
		return 0;
	}

	function addTrainStation(town, station, length) {
		myRailStationTownList.SetValue(town, station);
		if (length > 7)
			length = 77;	// means 7x2 station, largest supported now
		myRailStationLengthList.SetValue(town, length);
	}

	function BuildTrainStation(town) {
		Debug("Building train station...");
		// Find empty block of squares as close to target as possible
		local target = AITown.GetLocation(town);
		local spotFound = false;
		local curRange = 1;
		local length = abs(sqrt(AITown.GetPopulation(town)/100));
		local maxRange = 6 + length; //TODO check value correctness (coverage + size of something?)
		//local width = abs(length / 7) + 1;
		// length = abs(length % 7) + 1;
		// TEMP till support for bigger
		if (length > 7)
			length = 7;
		local width = (length == 7 ? 2 : 1);
		// TEMP
		//length = 7;
		//width = 2;

		// you need changing tracks for multiple platforms
		local totalLength = length + width;
		local area = AITileList();
		Debug("building station (lxb) " + length + " x " + width);

		while (curRange <= maxRange) {
			//Debug("looking with range " + curRange);
			area.AddRectangle(target + AIMap.GetTileIndex(curRange, curRange), target + AIMap.GetTileIndex(-curRange, -curRange));
			area.Valuate(AITile.GetCargoAcceptance, ::PASSENGER_ID, 1, 1, stationRadius);
			area.RemoveBelowValue(8);
			area.Valuate(AITile.GetCargoProduction, ::PASSENGER_ID, 1, 1, stationRadius);
			area.RemoveValue(0);
			area.Sort(AIAbstractList.SORT_BY_VALUE, false);
			area.Valuate(stationLocationValuator, totalLength, width);
			area.RemoveValue(0);
			if (area.Count()) {
				Debug("Found " + area.Count() + " possible options");
				for (local t = area.Begin(); area.HasNext(); t = area.Next()) {
					if (area.GetValue(t) == AIRail.RAILTRACK_NE_SW) {
						if (!AITile.LevelTiles(t, t + AIMap.GetTileIndex(length + width, width)))
							Error(AIError.GetLastErrorString());
					} else if (area.GetValue(t) == AIRail.RAILTRACK_NW_SE) {
						if (!AITile.LevelTiles(t, t + AIMap.GetTileIndex(width, length + width)))
							Error(AIError.GetLastErrorString());
					}
					if (AIRail.BuildRailStation(t, area.GetValue(t), width, length, AIStation.STATION_JOIN_ADJACENT)) {
						// TODO: support for wider stations and having exits on the other side
						if (area.GetValue(t) == AIRail.RAILTRACK_NE_SW) {
							AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length, 0), AIRail.RAILTRACK_NE_SW);
							AIRail.BuildSignal(t + AIMap.GetTileIndex(length, 0), t + AIMap.GetTileIndex(length - 1, 0), AIRail.SIGNALTYPE_PBS);
							if (width == 2) {
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length, 1), AIRail.RAILTRACK_NE_SW);
								AIRail.BuildSignal(t + AIMap.GetTileIndex(length, 1), t + AIMap.GetTileIndex(length - 1, 1), AIRail.SIGNALTYPE_PBS);

								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length + 1, 0), AIRail.RAILTRACK_NE_SW);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length + 1, 1), AIRail.RAILTRACK_NE_SW);
								// changing cross
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length + 1, 0), AIRail.RAILTRACK_SW_SE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length + 1, 0), AIRail.RAILTRACK_NE_SE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length + 1, 1), AIRail.RAILTRACK_NW_NE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(length + 1, 1), AIRail.RAILTRACK_NW_SW);
							}/* else {
								Error("Stations wider then 2 not yet supported");
							}*/								
						} else if (area.GetValue(t) == AIRail.RAILTRACK_NW_SE) {
							AIRail.BuildRailTrack(t + AIMap.GetTileIndex(0, length), AIRail.RAILTRACK_NW_SE);
							AIRail.BuildSignal(t + AIMap.GetTileIndex(0, length), t + AIMap.GetTileIndex(0, length - 1), AIRail.SIGNALTYPE_PBS);
							if (width == 2) {
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(1, length), AIRail.RAILTRACK_NW_SE);
								AIRail.BuildSignal(t + AIMap.GetTileIndex(1, length), t + AIMap.GetTileIndex(1, length - 1), AIRail.SIGNALTYPE_PBS);

								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(0, length + 1), AIRail.RAILTRACK_NW_SE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(1, length + 1), AIRail.RAILTRACK_NW_SE);
								//cross
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(0, length + 1), AIRail.RAILTRACK_NW_SW);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(0, length + 1), AIRail.RAILTRACK_SW_SE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(1, length + 1), AIRail.RAILTRACK_NE_SE);
								AIRail.BuildRailTrack(t + AIMap.GetTileIndex(1, length + 1), AIRail.RAILTRACK_NW_NE);
							}/* else {
								Error("Stations wider then 2 not yet supported");
							}*/								
						}
						return t;
					}
					// TODO: if err == area not clear: demolish?
					//Debug("Error building passenger station " + AIError.GetLastErrorString());
					//AISign.BuildSign(t, AIError.GetLastErrorString());
				}
				curRange++;
			} else {
				curRange++;
			}
		}
		Debug("No possible station location found");
		return null;
	}

	/**
	 * upgrades small station to bigger one
	 * also reroutes trains to new station if needed
	 */
	function UpgradeTrainStation(town) {}
}

