/**
 *    This file is part of OtviAI.
 *
 *    OtviAI is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    OtviAI is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with OtviAI.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright 2008-2010 Otto Visser.
 * Suggestions/comments and bugs can go to: otviAI@otvi.nl
 */

class OtviTransport {
	/**
	 * Returns whether anything can be done with this transport type
	 * Returns boolean
	 */
	function isOperational();

	/**
	 * Returns minimum amount of money needed to establish a new route
	 * default distance used for calculation is 100
	 * Returns max int if expansion not possible
	 */
	function getExpansionCost(distance = 100);

	/**
	 * Returns minimum amount of money needed to enhance the last build route
	 * i.e.: add more vehicles to the route
	 * returns max int if enhancing is not possible
	 */
	function getEnhanceCost();

	/**
	 * returns whether there is a route that could use enhancement at the moment
	 */
	function canEnhance();

	/**
	 * Returns transport preference
	 * Lower is more preferable
	 */
	function getTownTransportCost(town1, town2);
	function getIndustryTransportCost(producer, endLocation, cargoID);

	/**
	 * Updates engines
	 */
	function updateEngines();

	/**
	 * Connects two given towns and starts transporting passengers (and mail if possible)
	 * Returns: true iff succesful, false otherwise
	 */
	function connect2Towns(town1, town2);

	/**
	 * Connects industry to given town or industry and starts transporting given cargo
	 */
	function connectIndustryToIndustry(supplier, receiver, cargoID);
	function connectIndustryToTown(supplier, town, cargoID);
	
	/**
	 * sorts based on expansion cost
	 */
	function expansionSort(a, b) {
		if (a.getExpansionCost() > b.getExpansionCost())
			return 1;
		else if (a.getExpansionCost() < b.getExpansionCost())
			return -1;
		return 0;
	}

	/**
	 * sorts based on expansion cost
	 */
	function enhanceSort(a, b) {
		if (a.getEnhanceCost() > b.getEnhanceCost())
			return 1;
		else if (a.getEnhanceCost() < b.getEnhanceCost())
			return -1;
		return 0;
	}

	function townTransportSort(a, b) {
		local costa = a.getTownTransportCost(::startLoc, ::endLoc);
		local costb = b.getTownTransportCost(::startLoc, ::endLoc);

		if (costa > costb)
			return 1;
		else if (costa < costb)
			return -1;
		return 0;
	}

	function industryTransportSort(a, b) {
		local costa = a.getIndustryTransportCost(::startIndustry, ::endLoc, ::cargo_ID);
		local costb = b.getIndustryTransportCost(::startIndustry, ::endLoc, ::cargo_ID);

		if (costa > costb)
			return 1;
		else if (costa < costb)
			return -1;
		return 0;
	}
}

